
// ----------------------------------------------------------------------------
//   LCD 
// ----------------------------------------------------------------------------
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>


#define FOSC 4096000// Clock Speed
#define BAUD 9600
#define MYUBRR FOSC/16/BAUD-1


#define RXBUFFERLEN 200

//Globale Var:
typedef struct{
  volatile unsigned char enablePattern; /* 0x02 fr oben  0x01 fr unten*/
	volatile unsigned char readptr; 	
	volatile unsigned char writeptr; 	
  unsigned char          rxbuffer[RXBUFFERLEN]; //reichlich Platz
}Tglob;

Tglob g;



//enablePattern muss fuer folgende Funktionen richtig gesetzt sein
void clearLcd(void);
void home1Lcd(void); //Set cursor at start of Line 1
void home2Lcd(void); //Set cursor at start of Line 2



//######################################################################################


/*
     B1  LCD enable Z12 (upper)
     B0  LCD enable Z34 (lower)
*/


#define LCD_UPPER g.enablePattern = 0x02
#define LCD_LOWER g.enablePattern = 0x01

#define  LCD_ENABLE    PORTB = ((PORTB & (~0x03)) | g.enablePattern)
#define  LCD_DISABLE   PORTB &= (~0x03)
#define  LCD_DATA_L    PORTC
#define  LCD_DATA_H    PORTD
#define  LCD_RS_ON     PORTD |= 0x04    //SET PD2 
#define  LCD_RS_OFF    PORTD &= (~0x04) //CLR PD2
//DDRx: Gesetztes Bit = Ausgang
#define  LCD_READ      {DDRC = 0x00; DDRD = 0x0e; PORTD |=0x08; /*set PD3 */}  
#define  LCD_WRITE     {PORTD &= (~0x08); DDRC = 0x0f; DDRD = 0xfe; PORTD &= (~0x08);}
#define  BUSYBIT       (PIND & 0x80)



//--------------------------------------------------------
ISR(USART_RXC_vect){//RS232 Daten Empfang
  g.rxbuffer[g.writeptr] = UDR;
  g.writeptr++;
  if (g.writeptr >= RXBUFFERLEN){
  	g.writeptr = 0;
  }
}


// ---------------------------------------------------------------------------
void delay_ms(int ms){
unsigned long n;
int i;
  for (i=0; i<ms; i++){
    for (n=0; n<510; n++){ //Mit -O3 Optimierung dauert 1 Durchlauf 11 Takte
        asm volatile ("push	r31");
        asm volatile ("nop");
        asm volatile ("nop");
        asm volatile ("nop");
        asm volatile ("nop");
        asm volatile ("pop	r31");
    }           
  }
}

//--------------------------------------------------------
void USART_Transmit( unsigned char data ){
//--------------------------------------------------------
	/* Wait for empty transmit buffer */
	while ( !( UCSRA & (1<<UDRE)) )
	;
	/* Put data into buffer, sends the data */
	UDR = data;
}
// ---------------------------------------------------------------------------
unsigned char nextChar(void){
  unsigned char c;	
  
  while (g.readptr == g.writeptr){
    ;
  }
	  
  c = g.rxbuffer[g.readptr++];
  if (g.readptr >= RXBUFFERLEN){
  	g.readptr = 0;
  }
  USART_Transmit(c); //ECHO
  return c;
}

// ---------------------------------------------------------------------------
/*! Send byte to display.
    Depending on the value of LCD_RS the byte is written to display(LCD_RS=1) or register (LCD_RS=0)
   \param b is the byte to be send.
*/// -------------------------------------------------------------------------
void lcdWrite(unsigned char b) {
unsigned char busy;
  
//      asm volatile ("nop"); // 1 NOP bei 8MHz braucht etwa  125ns.  t_AS: 40ns min  #BATRON, page 57
      LCD_WRITE;
      asm volatile ("nop");  
      LCD_ENABLE;
      asm volatile ("nop");  

      LCD_DATA_L = b & 0x0f;
      LCD_DATA_H = (LCD_DATA_H & 0x0f) |  (b & 0xf0) ;
      asm volatile ("nop");
      asm volatile ("nop");
      asm volatile ("nop");  
      asm volatile ("nop");  
      //t_DSW: 80ns min.
      //t_H: 10ns min
      //t_CYC 500ns min
      LCD_DISABLE;
      asm volatile ("nop");
      asm volatile ("nop");
      asm volatile ("nop");  
    
     
      //Busy-Flag prfen - warten, bis nicht mehr gesetzt
      LCD_READ;
      do{
        asm volatile ("nop");
        asm volatile ("nop");  
        asm volatile ("nop");
        LCD_ENABLE;
        asm volatile ("nop");  
        asm volatile ("nop");
        asm volatile ("nop");
        busy = BUSYBIT;
        LCD_DISABLE;
      }while (busy ); //Warte bis BUSY-Flag gelscht
      
      LCD_WRITE;
      asm volatile ("nop");  
}


// ---------------------------------------------------------------------------
void cmd2Lcd(unsigned char cmd) {
  LCD_RS_OFF;
  lcdWrite(cmd);
  LCD_RS_ON;
}
// ---------------------------------------------------------------------------
/*! Clear LCD and set cursor home
   (HD 44780 based displays only)
*/// -------------------------------------------------------------------------

void clearLcd(void) {
  cmd2Lcd(0x01);
}
// ---------------------------------------------------------------------------
/*! Set cursor home
   (HD 44780 based displays only)
*/// -------------------------------------------------------------------------
void home1Lcd(void) {
  cmd2Lcd(0x02);
}
// ---------------------------------------------------------------------------
/*! Set cursor at start of Line 2
   (HD 44780 based displays only)
*/// -------------------------------------------------------------------------
void home2Lcd(void) {
  cmd2Lcd(192);
}


		

// ---------------------------------------------------------------------------
/*! Initialize display (2 Lines, 8 Bit, no shift, increment)
   (HD 44780 based displays only)
*/// -------------------------------------------------------------------------
void lcdInitHalf(void) {

  LCD_RS_OFF;
  LCD_ENABLE;

  //Delay >15ms
  delay_ms(20);
  
  lcdWrite(0x30); // DB5=1 DB4=1        # see BATRON-Specification for HD44780, page 58
  delay_ms(15);    // wait more than 4.1ms

  lcdWrite(0x30); // DB5=1 DB4=1
  delay_ms(15);    // wait more than 100us

  lcdWrite(0x30);
  delay_ms(15);	   //just to be sure...

  lcdWrite(0x38); //2 Lines, 8Bit Data,  
  delay_ms(1);    //just for paranoia
  lcdWrite(0x38); //2 Lines, 8Bit Data,  
  delay_ms(1);    //just for paranoia

  lcdWrite(0x0c); // Display on, Cursor off
  delay_ms(1);    //just for paranoia  
/*
  lcdWrite(0x01); // Display clear
  delay_ms(3);
*/
  lcdWrite(0x06); // Entry mode; no shift
  delay_ms(2);
  
  LCD_RS_ON;
}
// ----- Initialisiere LCD -----------------------------------------------------
void lcdInit(void) {
    LCD_UPPER;
    lcdInitHalf();

    LCD_LOWER;
    lcdInitHalf();

    LCD_UPPER;
}
//----- String senden ----------------------------------------------------------
void sendStringToLCD(char *str)
{
  unsigned char i;

  //Zeichen fuer Zeichen senden (ohne abschliessende '\0')
  i = 0;
  while (str[i] !=0)
  {
    lcdWrite(str[i]);
    i++;
    // mehr als 27 Zeichen werden nie an das Display gesendet
    if (i > 27){
      break;
    }
  }
}



//--------------------------------------------------------
void sendstringRS232(unsigned char *text){
//--------------------------------------------------------
  while(*text){
    USART_Transmit(*text++);
  }
}



//----- Schreiben aufs Display  --------------------------------------------------
void sendMessageToLCD(void)
{

/*Spezialzeichen im String:
  |o  = Oberer Display-Controller (Zeile1+2)
  |u  = Unterer Display-Controller (Zeile 3+4)
  |*  = ClearLCD (auf gerade eingestelltem Controller)
  |1  = Home 1. Zeile 
  |2  = Home 2. Zeile
  |3  = Home 3. Zeile
  |4  = Home 4. Zeile
  |Gyx= GotoYX x:1..4  y:@..Z
  |C  = Cursor Ein (auf gerade eingestelltem Controller) 
  |B  = Cursor Ein + Blinken(auf gerade eingestelltem Controller) 
  |c  = Cursor Aus (auf gerade eingestelltem Controller) 
  Die HOME-Befehle schalten auch die Displayhlfte um
*/
static unsigned char c;	
static unsigned char offset;


  //Zeichen fuer Zeichen senden
  c = nextChar();
  if ('|'==c){ //Steuerkommando
//    sendstringRS232("<CMD>");

  	c = nextChar();
  	switch(c){
  	  case 'G':  
    		c = nextChar(); //Zeile (Y)
    		switch(c){
    		  default: LCD_UPPER; offset=0;  break;		
    		  case '2':LCD_UPPER; offset=64; break;		
    		  case '3':LCD_LOWER; offset=0;  break;		
    		  case '4':LCD_LOWER; offset=64; break;		
    		}
    		c = nextChar() - 64; //Spalte (X)
    		cmd2Lcd(c + offset +0x80); //GotoYX 	      LCD_SET_DDRAM + x + (y * 64));	
  		break;
  	  case 'C': //Cursor EIN
    		cmd2Lcd(0x0E);
  		break;		
  	  case 'B': //Cursor EIN + Blinken
    		cmd2Lcd(0x0F);
  		break;		
  	  case 'c': //Cursor AUS
    		cmd2Lcd(0x0C);
  		break;		
  	  case 'o': LCD_UPPER; break;		
  	  case 'u': LCD_LOWER; break;		
  	  case '*': //CLEAR DISPLAY
    		cmd2Lcd(0x01);
  		break;
  	  case '1':
    		LCD_UPPER;
    		cmd2Lcd(2);//Home 1
  		break;
  	  case '2':
    		LCD_UPPER;
    		cmd2Lcd(192);//Home 2
  		break;
  	  case '3':
    		LCD_LOWER;
    		cmd2Lcd(2);//Home 1
  		break;
  	  case '4':
    		LCD_LOWER;
    		cmd2Lcd(192);//Home 2
  		break;
  	}		
  }else{ //Normales Zeichen
  	// ggf. Sonderzeichen umsetzen
  	if (c >= 0xB0){
  		if (c == 0xE4) c=0xE1;		// 
  		if (c == 0xF6) c=0xEF;		// 
  		if (c == 0xFC) c=0xF5;		// 
  		if (c == 0xDF) c=0xE2;		// 
  		if (c == 0xB0) c=0xDF;		//  (Grad)
  	}
  	lcdWrite(c);
  }
}


//----- RS232 initialisieren --------------------------------------------------


void USART_Init( unsigned int ubrr)
{
  /* Set baud rate */
  UBRRH = (unsigned char)(ubrr>>8);
  UBRRL = (unsigned char)ubrr;
  
  UCSRA=0; //Clear U2X
  
  /* Enable receiver and transmitter and RX-Interrupts*/
  UCSRB = (1<<RXEN)|(1<<TXEN)| (1<<RXCIE);
  
  /* Set frame format: 8data, 1stop bit */
  UCSRC = (1<<URSEL)|(3<<UCSZ0);
}


//----- SPI initialisieren --------------------------------------------------

char SPI_SlaveReceive(void)
{
/* Wait for reception complete */
while(!(SPSR & (1<<SPIF)))
;
/* Return data register */
return SPDR;
}


//----------------------------------------------------------------------------
int main( void )
{
unsigned char *ram;
unsigned char i;

  cli();
  //Gesetztes Bit: Ausgang
  DDRB = 0xff;
  DDRC = 0x0f;
  DDRD = 0xfe;

  /* Enable SPI */
  //SPCR = (1<<SPE)|(1<<SPIE);


  //Alle globalen Variablen auf Null
  ram=(unsigned char *)&g;
  for (i=0; i< sizeof(Tglob); ram[i++]=0);   


  USART_Init ( MYUBRR );

  lcdInit();

  clearLcd();
  home1Lcd();
  sendStringToLCD("Dietmar Schroeder 12/2008");
  home2Lcd();

  sei();

  sendstringRS232("\r\nSerial Display by Dietmar Schroeder 12/2008\r\n\r\n");
  sendstringRS232(
  "Spezialzeichen im String:\r\n"
  "  |o  = Oberer Display-Controller (Zeile1+2)\r\n"
  "  |u  = Unterer Display-Controller (Zeile 3+4)\r\n"
  "  |*  = ClearLCD (auf gerade eingestelltem Controller)\r\n"
  "  |1  = Home 1. Zeile\r\n"
  "  |2  = Home 2. Zeile\r\n"
  "  |3  = Home 3. Zeile\r\n"
  "  |4  = Home 4. Zeile\r\n"
  "  |Gyx= GotoYX x:1..4  y:@..Z\r\n"
  "  |C  = Cursor Ein (auf gerade eingestelltem Controller)\r\n"
  "  |B  = Cursor Ein + Blinken(auf gerade eingestelltem Controller)\r\n"
  "  |c  = Cursor Aus (auf gerade eingestelltem Controller)\r\n" 
  "  Die HOME-Befehle schalten auch die Displayhaelfte um\r\n"
  "READY.\r\n"
  );



	LCD_UPPER;
  home2Lcd();
  sendStringToLCD("READY.");

  while(1){ 
    sendMessageToLCD();
  }
  return 0;
}

	
